/* This code comes from the MAGIC library at http://www.cs.unc.edu/~eberly/ */

//---------------------------------------------------------------------------
// MinPointLine
//
// Compute the closest point on a line segment to a specified point.  The
// idea is to project the point onto the line containing the line segment.
// If the projected point is not on the line segment, then one of the line
// segment endpoints is closest to the specified point.  If the projected
// point is on the line segment, then that line segment point is closest.
//
//---------------------------------------------------------------------------
// MinLineLine
//
// This code computes the closest points on two line segments L0(s) = b0+s*m0,
// 0 <= s <= 1, and L1(t) = b1+t*m1, 0 <= t <= 1.  Given these lines, the
// procedure MinLineLine returns the s and t values for the closest points.
//
// In calculus terms, the goal is to minimize the squared-distance function
// Q(s,t) = Dot(L0(s)-L1(t),L0(s)-L1(t)) over the unit square 0 <= s <= 1
// and 0 <= t <= 1.  This function is
//
//   Q(s,t) = A*s*s + 2*B*s*t + C*t*t + 2*D*s + 2*E*t + F
//
// where A = Dot(m0,m0), B = -Dot(m0,m1), C = Dot(m1,m1), D = Dot(m0,b0-b1),
// E = -Dot(m1,b0-b1), and F = Dot(b0-b1,b0-b1).
//
// As a function over all of s and t, Q is nonnegative and has an absolute
// minimum occurring at grad(Q) = 2*(A*s+B*t+D,B*s+C*t+E) = (0,0).  This
// is a system of two equations in two unknowns.  The system has a unique
// solution when A*C-B*B != 0.  In this case, if the solution (s,t) lies
// in the unit square, then the closest points on the line segments are
// both interior points of their respective segments.  If (s,t) lies outside
// the unit square, then the minimum of Q occurs on the boundary of the
// unit square.  To find the correct boundary, consider the following
// diagram.
//
//          t
//          |         |
//     4    |    3    |    2
//          |         |
// ---------+---------+---------
//          |         |
//          |         |
//     5    |    0    |    1
//          |         |
//          |         |
// ---------+---------+--------- s
//          |         |
//     6    |    7    |    8
//          |         |
//
// The central square labeled region 0 is the domain of Q, 0 <= s <= 1 and
// 0 <= t <= 1.  As stated earlier, tf (s,t) lies in region 0, then the two
// closest points are interior points of their respective segments.  If
// (s,t) lies in region 1, then since the level curves of Q are ellipses
// (and therefore convex), the minimum of Q lies on s=1 for 0 <= t <= 1 at
// the point where a level curve of Q is tangent to curve s=1.  Similarly,
// if (s,t) is in region 3, then the minimum is on t=1 for 0 <= s <= 1.  If
// (s,t) is in region 5, the minimum is on s=0 for 0 <= t <= 1.  If (s,t)
// is in region 7, the minimum is on t=0 for 0 <= s <= 1.
//
// If (s,t) is in region 2, it is possible that the level curve of Q which
// is tangent somewhere to the unit square touches either on edge t=1 or
// edge s=1.  Or it is possible that the level curve grazes the corner
// (1,1), in which case this point is where the minimum occurs.  To 
// distinguish these three cases, it is sufficient to look at the partial
// derivatives of Q at (1,1).  Let (Q_s,Q_t) = grad(Q) be the partial
// derivatives.  If Q_s(1,1) > 0, then Q decreases as you move along t=1
// for decreasing s.  This forces the minimum of Q to occur on edge t=1.
// If Q_t(1,1) > 0, then Q decreases as you move along s=1 for decreasing
// t.  This forces the minimum of Q to occur on edge s=1.  If neither of
// Q_s(1,1) and Q_t(1,1) is positive, then the minimum occurs at (1,1).
// Similar arguments apply to regions 4, 6, and 8.
//
// For the case when A*C-B*B = 0, the global minimum of Q occurs along
// an entire line, not at a single isolated point.  Geometrically the
// line segments are parallel, so there are many points on both lines
// which attain the minimum distance between the lines.  The algorithm I
// use is to project Line1(t) onto Line0(s).  If the projections are
// disjoint, then the minimum occurs at two endpoints on the line
// segments.  If the projections overlap, then I choose an s and t in
// the overlap to get a pair of closest points.
//
//---------------------------------------------------------------------------
// MinPointTriangle
//
// This code computes the closest point of a triangle Tri(s,t) = b+s*e0+t*e1,
// 0 <= s <= 1, 0 <= t <= 1, and 0 <= s+t <= 1, to a specified point P.
//
// In calculus terms, the goal is to minimize the squared-distance function
// Q(s,t) = Dot(Tri(s,t)-p,Tri(s,t)-p) over the triangular domain
// 0 <= s <= 1, 0 <= t <= 1, 0 <= s+t <= 1.  This function is
//
//   Q(s,t) = A*s*s + 2*B*s*t + C*t*t + 2*D*s + 2*E*t + F
//
// where A = Dot(e0,e0), B = Dot(e0,e1), C = Dot(e1,e1), D = Dot(e0,b-P),
// E = Dot(e1,b-P), and F = Dot(b-P,b-P).
//
// As a function over all of s and t, Q is nonnegative and has an absolute
// minimum occurring at grad(Q) = 2*(A*s+B*t+D,B*s+C*t+E) = (0,0).  This
// is a system of two equations in two unknowns.  For non-degenerate
// triangles, A*C-B*B != 0.  If the solution (s,t) lies in the triangular
// domain, then the closest point on the triangle is interior.  If (s,t)
// lies outside the triangular domain, then the minimum of Q occurs on the
// boundary of the domain.  To find the correct boundary, consider the
// following diagram.
//
//          t
//    \   2 |                
//      \   |                
//        \ |                
//          +                    
//          |\        1
//          |  \       
//     3    |    \           
//          |  0   \   
//          |        \ 
// ---------+---------+--------- s
//          |           \
//     4    |    5        \  6
//          |               \
//
// If (s,t) is in region 0, the closest triangle point is an interior point.
// If (s,t) is in region 1, then the minimum must occur at a point on the
// segment in the first quadrant separating regions 0 and 1.  If (s,t) is
// in region 3, the minimum occurs at a point on the segment separating
// regions 0 and 3.  Similarly if (s,t) is in region 5, then the minimum
// must occur on the segment separating regions 0 and 5.  If (s,t) is in
// region 2, the the minimum may occur either on the segment between
// region 0 and 3 or on the segment between region 0 and 1.  A gradient
// test at (0,1) will determine which is the case.  Similary if (s,t) is
// in region 6, the minimum may occur either on the segment between
// region 0 and 5 or on the segment between region 0 and 1.  A gradient
// test at (1,0) will determine which is the case.  Finally, if (s,t) is
// in region 4, the minimum may occur either on the segment between
// region 0 and 3 or on the segment between region 0 and 5.  Again, a
// gradient test will determine which is the case.
//
//---------------------------------------------------------------------------
// MinLineTriangle
//
// This code computes the closest points of a line L(r) = a0+r*a1 and a
// triangle Tri(s,t) = b0+s*b1+t*b2, where 0 <= r <= 1 and 0 <= s <= 1,
// 0 <= t <= 1, and 0 <= s+t <= 1.
//
// In calculus terms, the goal is to minimize the squared-distance function
// Q(r,s,t) = Dot(L(r)-Tri(s,t),L(r)-Tri(s,t)) over the prism domain
// 0 <= r <= 1, 0 <= s <= 1, 0 <= t <= 1, 0 <= s+t <= 1.  This function is
//
//   Q(r,s,t) = [r s t] A [r s t] + 2 B [r s t] + C
//
// where A is a 3x3 symmetric matrix, B is a 3x1 column vector, and C is
// a constant.  The entries of A and B are various dot products derived
// from the vectors a0, a1, b0, b1, and b2.
//
// The analysis is similar to that of MinPointTriangle.  The (s,t) domain
// is partitioned similarly, but the prism is obtained by extruding the
// domain in the r-direction.  The three cases for r are: r < 0, 0 <= r <= 1,
// and r > 1.  For each case the partitioning in (s,t) is identical to that
// of MinPointTriangle.  The region numbers for r < 0 are appended with an
// 'm' and the region numbers for r > 1 are appended with a 'p'.
//
//---------------------------------------------------------------------------
// MinTriangleTriangle
//
// The quadratic form representing the squared distance between two planes
// never has an isolated global minimum.  The generic case is for it to
// have an entire line of zeros (the line of intersection of the two
// planes).  Therefore, it is sufficient to compare edges of each triangle
// with the entire other triangle, looking for the minimum distance.
//---------------------------------------------------------------------------
// MinPointRectoid and MinLineRectoid are similar in construction to
// MinPointTriangle and MinLineTriangle, the only difference being that
// for rectoids, 0 <= s <= 1 and 0 <= t <= 1, whereas for triangles,
// additionally 0 <= s+1 <= 1.  The coding differences are just in what
// cases to consider.  MinTriangleRectoid and MinRectoidRectoid simply
// decompose the rectoids into triangles and call MinTriangleTriangle.
//---------------------------------------------------------------------------

#include <math.h>
#include "geomdist.h"

//---------------------------------------------------------------------------
static double Dot (const Point3& p, const Point3& q)
{
	return p.x*q.x+p.y*q.y+p.z*q.z;
}
//---------------------------------------------------------------------------
double MinPointLine (const Point3& p, const Line3& line,
					 double& t)
{
	// Find t so that the point t*m+b on the line segment (0 <= t <= 1)
	// is closest to point p.

	Point3 diff = { p.x-line.b.x, p.y-line.b.y, p.z-line.b.z };
	t = Dot(line.m,diff)/Dot(line.m,line.m);
	if ( t < 0.0 ) t = 0.0; else if ( t > 1.0 ) t = 1.0;

	diff.x -= t*line.m.x;
	diff.y -= t*line.m.y;
	diff.z -= t*line.m.z;
	return sqrt(Dot(diff,diff));
}
//---------------------------------------------------------------------------
double MinLineLine (const Line3& line0, const Line3& line1,
					double& s, double& t)
{
	const double tolerance = 1e-05;

	Point3 diff =
		{ line0.b.x-line1.b.x, line0.b.y-line1.b.y, line0.b.z-line1.b.z };

	double A = Dot(line0.m,line0.m);
	double B = -Dot(line0.m,line1.m);
	double C = Dot(line1.m,line1.m);
	double D = Dot(line0.m,diff);
	double E = -Dot(line1.m,diff);
	double det = A*C-B*B;

	if ( fabs(det) >= tolerance )
	{
		// line segments are not parallel
		s = (B*E-C*D)/det;
		t = (B*D-A*E)/det;
		
		if ( s >= 0.0 )
		{
			if ( s <= 1.0 )
			{
				if ( t >= 0.0 )
				{
					if ( t <= 1.0 )  // region 0
					{
						// minimum at two interiors
					}
					else  // region 3
					{
						t = 1.0;
						s = -(B+D)/A;
						if ( s < 0.0 ) s = 0.0; else if ( s > 1.0 ) s = 1.0;
					}
				}
				else  // region 7
				{
					t = 0.0;
					s = -D/A;
					if ( s < 0.0 ) s = 0.0; else if ( s > 1.0 ) s = 1.0;
				}
			}
			else
			{
				if ( t >= 0.0 )
				{
					if ( t <= 1.0 )  // region 1
					{
						s = 1.0;
						t = -(B+E)/C;
						if ( t < 0.0 ) t = 0.0; else if ( t > 1.0 ) t = 1.0;
					}
					else  // region 2
					{
						if ( A+B+D > 0 )
						{
							t = 1.0;
							s = -(B+D)/A;
							if ( s < 0.0 ) s = 0.0;
						}
						else if ( B+C+E > 0 )
						{
							s = 1.0;
							t = -(B+E)/C;
							if ( t < 0.0 ) t = 0.0;
						}
						else
						{
							s = 1.0;
							t = 1.0;
						}
					}
				}
				else  // region 8
				{
					if ( A+D > 0 )
					{
						t = 0.0;
						s = -D/A;
						if ( s < 0.0 ) s = 0.0;
					}
					else if ( B+E < 0 )
					{
						s = 1.0;
						t = -(B+E)/C;
						if ( t > 1.0 ) t = 1.0;
					}
					else
					{
						s = 1.0;
						t = 0.0;
					}
				}
			}
		}
		else 
		{
			if ( t >= 0.0 )
			{
				if ( t <= 1.0 )  // region 5
				{
					s = 0.0;
					t = -E/C;
					if ( t < 0.0 ) t = 0.0; else if ( t > 1.0 ) t = 1.0;
				}
				else  // region 4
				{
					if ( B+D < 0 )
					{
						t = 1.0;
						s = -(B+D)/A;
						if ( s > 1.0 ) s = 1.0;
					}
					else if ( C+E > 0 )
					{
						s = 0.0;
						t = -E/C;
						if ( t < 0.0 ) t = 0.0;
					}
					else
					{
						s = 0.0;
						t = 1.0;
					}
				}
			}
			else   // region 6
			{
				if ( D < 0 )
				{
					t = 0.0;
					s = -D/A;
					if ( s > 1.0 ) s = 1.0;
				}
				else if ( E < 0 )
				{
					s = 0.0;
					t = -E/C;
					if ( t > 1.0 ) t = 1.0;
				}
				else
				{
					s = 0.0;
					t = 0.0;
				}
			}
		}
	}
	else
	{
		// line segments are parallel
		double t0, t1;
		if ( B > 0.0 )  // t0 > t1
		{
			t0 = -D/A;
			if ( t0 <= 0.0 )
			{
				s = 0.0;
				t = 0.0;
			}
			else if ( t0 <= 1.0 )
			{
				s = t0;
				t = 0.0;
			}
			else
			{
				t1 = -(B+D)/A;
				if ( t1 <= 1.0 )
				{
					s = 1.0;
					t = (t0-1.0)/(t0-t1);
				}
				else
				{
					s = 1.0;
					t = 1.0;
				}
			}
		}
		else  // t0 < t1
		{
			t1 = -(B+D)/A;
			if ( t1 <= 0.0 )
			{
				s = 0.0;
				t = 1.0;
			}
			else if ( t1 <= 1.0 )
			{
				s = t1;
				t = 1.0;
			}
			else
			{
				t0 = -D/A;
				if ( t0 <= 1.0 )
				{
					s = 1.0;
					t = (1.0-t0)/(t1-t0);
				}
				else
				{
					s = 1.0;
					t = 0.0;
				}
			}
		}
	}

	return sqrt(fabs(s*(A*s+B*t+2*D)+t*(B*s+C*t+2*E)+Dot(diff,diff)));
}
//---------------------------------------------------------------------------
double MinPointTriangle (const Point3& p, const Triangle3& tri,
						 double& s, double& t)
{
	Point3 diff = { tri.b.x-p.x, tri.b.y-p.y, tri.b.z-p.z };
	double A = Dot(tri.e0,tri.e0);
	double B = Dot(tri.e0,tri.e1);
	double C = Dot(tri.e1,tri.e1);
	double D = Dot(tri.e0,diff);
	double E = Dot(tri.e1,diff);
	double F = Dot(diff,diff);
	double det = A*C-B*B;  // assert:  det != 0 for triangles

	s = (B*E-C*D)/det;
	t = (B*D-A*E)/det;

	if ( s+t <= 1.0 )
	{
		if ( s < 0.0 )
		{
			if ( t < 0.0 )  // region 4
			{
				if ( D < 0 )
				{
					t = 0.0;
					s = -D/A;
					if ( s > 1.0 ) s = 1.0;
				}
				else if ( E < 0 )
				{
					s = 0.0;
					t = -E/C;
					if ( t > 1.0 ) t = 1.0;
				}
				else
				{
					s = 0.0;
					t = 0.0;
				}
			}
			else  // region 3
			{
				s = 0.0;
				t = -E/C;
				if ( t < 0.0 ) t = 0.0; else if ( t > 1.0 ) t = 1.0;
			}
		}
		else if ( t < 0.0 )  // region 5
		{
			t = 0.0;
			s = -D/A;
			if ( s < 0.0 ) s = 0.0; else if ( s > 1.0 ) s = 1.0;
		}
		else  // region 0
		{
			// minimum at interior point
		}
	}
	else
	{
		if ( s < 0.0 )  // region 2
		{
			if ( B-C+D-E < 0.0 )
			{
				s = -(B-C+D-E)/(A-2*B+C);
				if ( s < 0.0 ) s = 0.0; else if ( s > 1.0 ) s = 1.0;
				t = 1.0-s;
			}
			else if ( C+E > 0.0 )
			{
				s = 0.0;
				t = -E/C;
				if ( t < 0.0 ) t = 0.0; else if ( t > 1.0 ) t = 1.0;
			}
			else
			{
				s = 0.0;
				t = 1.0;
			}
		}
		else if ( t < 0.0 )  // region 6
		{
			if ( A-B+D-E > 0.0 )
			{
				t = (A-B+D-E)/(A-2*B+C);
				if ( t < 0.0 ) t = 0.0; else if ( t > 1.0 ) t = 1.0;
				s = 1.0-t;
			}
			else if ( A+D > 0.0 )
			{
				t = 0.0;
				s = -D/A;
				if ( s < 0.0 ) s = 0.0; else if ( s > 1.0 ) s = 1.0;
			}
			else
			{
				s = 1.0;
				t = 0.0;
			}
		}
		else  // region 1
		{
			s = -(B-C+D-E)/(A-2*B+C);
			if ( s < 0.0 ) s = 0.0; else if ( s > 1.0 ) s = 1.0;
			t = 1.0-s;
		}
	}

	return sqrt(fabs(s*(A*s+B*t+2*D)+t*(B*s+C*t+2*E)+F));
}
//---------------------------------------------------------------------------
double MinLineTriangle (const Line3& line, const Triangle3& tri,
						double& r, double& s, double& t)
{
	Point3 diff = { tri.b.x-line.b.x, tri.b.y-line.b.y, tri.b.z-line.b.z };
	double A00 = +Dot(line.m,line.m);
	double A01 = -Dot(line.m,tri.e0);
	double A02 = -Dot(line.m,tri.e1);
	double A11 = +Dot(tri.e0,tri.e0);
	double A12 = +Dot(tri.e0,tri.e1);
	double A22 = +Dot(tri.e1,tri.e1);
	double B0  = -Dot(diff,line.m);
	double B1  = +Dot(diff,tri.e0);
	double B2  = +Dot(diff,tri.e1);
	double cof00 = A11*A22-A12*A12;
	double cof01 = A02*A12-A01*A22;
	double cof02 = A01*A12-A02*A11;
	double det = A00*cof00+A01*cof01+A02*cof02;

	const double tolerance = 1e-05;
	Line3 triline;
	Point3 pt;
	double min, min0, r0, s0, t0;

	if ( fabs(det) >= tolerance )
	{
		double cof11 = A00*A22-A02*A02;
		double cof12 = A02*A01-A00*A12;
		double cof22 = A00*A11-A01*A01;
		double rhs0 = -B0/det;
		double rhs1 = -B1/det;
		double rhs2 = -B2/det;

		r = cof00*rhs0+cof01*rhs1+cof02*rhs2;
		s = cof01*rhs0+cof11*rhs1+cof12*rhs2;
		t = cof02*rhs0+cof12*rhs1+cof22*rhs2;

		if ( r < 0.0 )
		{
			if ( s+t <= 1.0 )
			{
				if ( s < 0.0 )
				{
					if ( t < 0.0 )  // region 4m
					{
						// min on face s=0 or t=0 or r=0
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e1.x;
						triline.m.y = tri.e1.y;
						triline.m.z = tri.e1.z;
						min = MinLineLine(line,triline,r,t);
						s = 0.0;
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e0.x;
						triline.m.y = tri.e0.y;
						triline.m.z = tri.e0.z;
						min0 = MinLineLine(line,triline,r0,s0);
						t0 = 0.0;
						if ( min0 < min )
						{
							min = min0;
							r = r0;
							s = s0;
							t = t0;
						}
						min0 = MinPointTriangle(line.b,tri,s0,t0);
						r0 = 0.0;
						if ( min0 < min )
						{
							min = min0;
							r = r0;
							s = s0;
							t = t0;
						}
					}
					else  // region 3m
					{
						// min on face s=0 or r=0
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e1.x;
						triline.m.y = tri.e1.y;
						triline.m.z = tri.e1.z;
						min = MinLineLine(line,triline,r,t);
						s = 0.0;
						min0 = MinPointTriangle(line.b,tri,s0,t0);
						r0 = 0.0;
						if ( min0 < min )
						{
							min = min0;
							r = r0;
							s = s0;
							t = t0;
						}
					}
				}
				else if ( t < 0.0 )  // region 5m
				{
					// min on face t=0 or r=0
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e0.x;
					triline.m.y = tri.e0.y;
					triline.m.z = tri.e0.z;
					min = MinLineLine(line,triline,r,s);
					t = 0.0;
					min0 = MinPointTriangle(line.b,tri,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 0m
				{
					// min face on r=0
					min = MinPointTriangle(line.b,tri,s,t);
					r = 0.0;
				}
			}
			else
			{
				if ( s < 0.0 )  // region 2m
				{
					// min on face s=0 or s+t=1 or r=0
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e1.x;
					triline.m.y = tri.e1.y;
					triline.m.z = tri.e1.z;
					min = MinLineLine(line,triline,r,t);
					s = 0.0;
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min0 = MinLineLine(line,triline,r0,t0);
					s0 = 1.0-t0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					min0 = MinPointTriangle(line.b,tri,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t < 0.0 )  // region 6m
				{
					// min on face t=0 or s+t=1 or r=0
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e0.x;
					triline.m.y = tri.e0.y;
					triline.m.z = tri.e0.z;
					min = MinLineLine(line,triline,r,s);
					t = 0.0;
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min0 = MinLineLine(line,triline,r0,t0);
					s0 = 1.0-t0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					min0 = MinPointTriangle(line.b,tri,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 1m
				{
					// min on face s+t=1 or r=0
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min = MinLineLine(line,triline,r,t);
					s = 1.0-t;
					min0 = MinPointTriangle(line.b,tri,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
		}
		else if ( r <= 1.0 )
		{
			if ( s+t <= 1.0 )
			{
				if ( s < 0.0 )
				{
					if ( t < 0.0 )  // region 4
					{
						// min on face s=0 or t=0
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e1.x;
						triline.m.y = tri.e1.y;
						triline.m.z = tri.e1.z;
						min = MinLineLine(line,triline,r,t);
						s = 0.0;
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e0.x;
						triline.m.y = tri.e0.y;
						triline.m.z = tri.e0.z;
						min0 = MinLineLine(line,triline,r0,s0);
						t0 = 0.0;
						if ( min0 < min )
						{
							min = min0;
							r = r0;
							s = s0;
							t = t0;
						}
					}
					else  // region 3
					{
						// min on face s=0
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e1.x;
						triline.m.y = tri.e1.y;
						triline.m.z = tri.e1.z;
						min = MinLineLine(line,triline,r,t);
						s = 0.0;
					}
				}
				else if ( t < 0.0 )  // region 5
				{
					// min on face t=0
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e0.x;
					triline.m.y = tri.e0.y;
					triline.m.z = tri.e0.z;
					min = MinLineLine(line,triline,r,s);
					t = 0.0;
				}
				else  // region 0
				{
					// global minimum is interior, done
					min = sqrt(fabs(r*(A00*r+A01*s+A02*t+2.0*B0)
						  +s*(A01*r+A11*s+A12*t+2.0*B1)
						  +t*(A02*r+A12*s+A22*t+2.0*B2)
						  +Dot(diff,diff)));
				}
			}
			else
			{
				if ( s < 0.0 )  // region 2
				{
					// min on face s=0 or s+t=1
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e1.x;
					triline.m.y = tri.e1.y;
					triline.m.z = tri.e1.z;
					min = MinLineLine(line,triline,r,t);
					s = 0.0;
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min0 = MinLineLine(line,triline,r0,t0);
					s0 = 1.0-t0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t < 0.0 )  // region 6
				{
					// min on face t=0 or s+t=1
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e0.x;
					triline.m.y = tri.e0.y;
					triline.m.z = tri.e0.z;
					min = MinLineLine(line,triline,r,s);
					t = 0.0;
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min0 = MinLineLine(line,triline,r0,t0);
					s0 = 1.0-t0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 1
				{
					// min on face s+t=1
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min = MinLineLine(line,triline,r,t);
					s = 1.0-t;
				}
			}
		}
		else  // r > 1.0
		{
			if ( s+t <= 1.0 )
			{
				if ( s < 0.0 )
				{
					if ( t < 0.0 )  // region 4p
					{
						// min on face s=0 or t=0 or r=1
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e1.x;
						triline.m.y = tri.e1.y;
						triline.m.z = tri.e1.z;
						min = MinLineLine(line,triline,r,t);
						s = 0.0;
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e0.x;
						triline.m.y = tri.e0.y;
						triline.m.z = tri.e0.z;
						min0 = MinLineLine(line,triline,r0,s0);
						t0 = 0.0;
						if ( min0 < min )
						{
							min = min0;
							r = r0;
							s = s0;
							t = t0;
						}
						pt.x = line.b.x+line.m.x;
						pt.y = line.b.y+line.m.y;
						pt.z = line.b.z+line.m.z;
						min0 = MinPointTriangle(pt,tri,s0,t0);
						r0 = 1.0;
						if ( min0 < min )
						{
							min = min0;
							r = r0;
							s = s0;
							t = t0;
						}
					}
					else  // region 3p
					{
						// min on face s=0 or r=1
						triline.b.x = tri.b.x;
						triline.b.y = tri.b.y;
						triline.b.z = tri.b.z;
						triline.m.x = tri.e1.x;
						triline.m.y = tri.e1.y;
						triline.m.z = tri.e1.z;
						min = MinLineLine(line,triline,r,t);
						s = 0.0;
						pt.x = line.b.x+line.m.x;
						pt.y = line.b.y+line.m.y;
						pt.z = line.b.z+line.m.z;
						min0 = MinPointTriangle(pt,tri,s0,t0);
						r0 = 1.0;
						if ( min0 < min )
						{
							min = min0;
							r = r0;
							s = s0;
							t = t0;
						}
					}
				}
				else if ( t < 0.0 )  // region 5p
				{
					// min on face t=0 or r=1
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e0.x;
					triline.m.y = tri.e0.y;
					triline.m.z = tri.e0.z;
					min = MinLineLine(line,triline,r,s);
					t = 0.0;
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointTriangle(pt,tri,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 0p
				{
					// min face on r=1
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min = MinPointTriangle(pt,tri,s,t);
					r = 1.0;
				}
			}
			else
			{
				if ( s < 0.0 )  // region 2p
				{
					// min on face s=0 or s+t=1 or r=1
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e1.x;
					triline.m.y = tri.e1.y;
					triline.m.z = tri.e1.z;
					min = MinLineLine(line,triline,r,t);
					s = 0.0;
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min0 = MinLineLine(line,triline,r0,t0);
					s0 = 1.0-t0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointTriangle(pt,tri,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t < 0.0 )  // region 6p
				{
					// min on face t=0 or s+t=1 or r=1
					triline.b.x = tri.b.x;
					triline.b.y = tri.b.y;
					triline.b.z = tri.b.z;
					triline.m.x = tri.e0.x;
					triline.m.y = tri.e0.y;
					triline.m.z = tri.e0.z;
					min = MinLineLine(line,triline,r,s);
					t = 0.0;
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min0 = MinLineLine(line,triline,r0,t0);
					s0 = 1.0-t0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointTriangle(pt,tri,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 1p
				{
					// min on face s+t=1 or r=1
					triline.b.x = tri.b.x+tri.e0.x;
					triline.b.y = tri.b.y+tri.e0.y;
					triline.b.z = tri.b.z+tri.e0.z;
					triline.m.x = tri.e1.x-tri.e0.x;
					triline.m.y = tri.e1.y-tri.e0.y;
					triline.m.z = tri.e1.z-tri.e0.z;
					min = MinLineLine(line,triline,r,t);
					s = 1.0-t;
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointTriangle(pt,tri,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
		}
	}
	else
	{
		// line and triangle are parallel
		triline.b.x = tri.b.x;
		triline.b.y = tri.b.y;
		triline.b.z = tri.b.z;
		triline.m.x = tri.e0.x;
		triline.m.y = tri.e0.y;
		triline.m.z = tri.e0.z;
		min = MinLineLine(line,triline,r,s);
		t = 0.0;

		triline.m.x = tri.e1.x;
		triline.m.y = tri.e1.y;
		triline.m.z = tri.e1.z;
		min0 = MinLineLine(line,triline,r0,t0);
		s0 = 0.0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}

		triline.b.x += tri.e0.x;
		triline.b.y += tri.e0.y;
		triline.b.z += tri.e0.z;
		triline.m.x -= tri.e0.x;
		triline.m.y -= tri.e0.y;
		triline.m.z -= tri.e0.z;
		min0 = MinLineLine(line,triline,r0,t0);
		s0 = 1.0-t0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}

		min0 = MinPointTriangle(line.b,tri,s0,t0);
		r0 = 0.0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}

		pt.x = line.b.x+line.m.x;
		pt.y = line.b.y+line.m.y;
		pt.z = line.b.z+line.m.z;
		min0 = MinPointTriangle(pt,tri,s0,t0);
		r0 = 1.0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}
	}

	return min;
}
//---------------------------------------------------------------------------
double MinTriangleTriangle (const Triangle3& tri0, const Triangle3& tri1,
							double& s, double& t, double& u, double& v)
{
	double s0, t0, u0, v0, min, min0;
	Line3 line;

	// compare edges of tri0 against all of tri1
	line.b.x = tri0.b.x;
	line.b.y = tri0.b.y;
	line.b.z = tri0.b.z;
	line.m.x = tri0.e0.x;
	line.m.y = tri0.e0.y;
	line.m.z = tri0.e0.z;
	min = MinLineTriangle(line,tri1,s,u,v);
	t = 0.0;

	line.m.x = tri0.e1.x;
	line.m.y = tri0.e1.y;
	line.m.z = tri0.e1.z;
	min0 = MinLineTriangle(line,tri1,t0,u0,v0);
	s0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x += tri0.e0.x;
	line.b.y += tri0.e0.y;
	line.b.z += tri0.e0.z;
	line.m.x -= tri0.e0.x;
	line.m.y -= tri0.e0.y;
	line.m.z -= tri0.e0.z;
	min0 = MinLineTriangle(line,tri1,t0,u0,v0);  // fix:  replace s0 with t0
	s0 = 1.0-t0;  // fix:  swap s0 and t0
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	// compare edges of tri1 against all of tri0
	line.b.x = tri1.b.x;
	line.b.y = tri1.b.y;
	line.b.z = tri1.b.z;
	line.m.x = tri1.e0.x;
	line.m.y = tri1.e0.y;
	line.m.z = tri1.e0.z;
	min0 = MinLineTriangle(line,tri0,u0,s0,t0);
	v0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.m.x = tri1.e1.x;
	line.m.y = tri1.e1.y;
	line.m.z = tri1.e1.z;
	min0 = MinLineTriangle(line,tri0,v0,s0,t0);
	u0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x += tri1.e0.x;
	line.b.y += tri1.e0.y;
	line.b.z += tri1.e0.z;
	line.m.x -= tri1.e0.x;
	line.m.y -= tri1.e0.y;
	line.m.z -= tri1.e0.z;
	min0 = MinLineTriangle(line,tri0,v0,s0,t0);  // fix: replace u0 with v0
	u0 = 1.0-v0;  // fix: swap u0 and v0
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	return min;
}
//---------------------------------------------------------------------------
double MinPointRectoid (const Point3& p, const Rectoid3& rect,
						double& s, double& t)
{
	Point3 diff = { rect.b.x-p.x, rect.b.y-p.y, rect.b.z-p.z };
	double A = Dot(rect.e0,rect.e0);
	double B = Dot(rect.e0,rect.e1);
	double C = Dot(rect.e1,rect.e1);
	double D = Dot(rect.e0,diff);
	double E = Dot(rect.e1,diff);
	double F = Dot(diff,diff);
	double det = A*C-B*B;  // assert:  det != 0 for rectoids

	s = (B*E-C*D)/det;
	t = (B*D-A*E)/det;

	if ( s < 0.0 )
	{
		if ( t < 0.0 )  // region 6
		{
			if ( D < 0.0 )
			{
				t = 0.0;
				s = -D/A;
				if ( s > 1.0 ) s = 1.0;
			}
			else if ( E < 0.0 )
			{
				s = 0.0;
				t = -E/C;
				if ( t > 1.0 ) t = 1.0;
			}
			else
			{
				s = 0.0;
				t = 0.0;
			}
		}
		else if ( t <= 1.0 )  // region 5
		{
			s = 0.0;
			t = -E/C;
			if ( t < 0.0 ) t = 0.0; else if ( t > 1.0 ) t = 1.0;
		}
		else  // region 4
		{
			if ( B+D < 0.0 )
			{
				t = 1.0;
				s = -(B+D)/A;
				if ( s > 1.0 ) s = 1.0;
			}
			else if ( C+E > 0.0 )
			{
				s = 0.0;
				t = -E/C;
				if ( t < 0.0 ) t = 0.0;
			}
			else
			{
				s = 0.0;
				t = 1.0;
			}
		}
	}
	else if ( s <= 1.0 )
	{
		if ( t < 0.0 )  // region 7
		{
			t = 0.0;
			s = -D/A;
			if ( s < 0.0 ) s = 0.0; else if ( s > 1.0 ) s = 1.0;
		}
		else if ( t <= 1.0 )  // region 0
		{
			// minimum at interior point
		}
		else  // region 3
		{
			t = 1.0;
			s = -(B+D)/A;
			if ( s < 0.0 ) s = 0.0; else if ( s > 1.0 ) s = 1.0;
		}
	}
	else
	{
		if ( t < 0.0 )  // region 8
		{
			if ( A+D > 0.0 )
			{
				t = 0.0;
				s = -D/A;
				if ( s < 0.0 ) s = 0.0;
			}
			else if ( B+E < 0.0 )
			{
				s = 1.0;
				t = -(B+E)/C;
				if ( t > 1.0 ) t = 1.0;
			}
			else
			{
				s = 1.0;
				t = 0.0;
			}
		}
		else if ( t <= 1.0 )  // region 1
		{
			s = 1.0;
			t = -(B+E)/C;
			if ( t < 0.0 ) t = 0.0; else if ( t > 1.0 ) t = 1.0;
		}
		else  // region 2
		{
			if ( A+B+D > 0.0 )
			{
				t = 1.0;
				s = -(B+D)/A;
				if ( s < 0.0 ) s = 0.0;
			}
			else if ( B+C+E > 0.0 )
			{
				s = 1.0;
				t = -(B+E)/C;
				if ( t < 0.0 ) t = 0.0;
			}
			else
			{
				s = 1.0;
				t = 1.0;
			}
		}
	}

	return sqrt(fabs(s*(A*s+B*t+2*D)+t*(B*s+C*t+2*E)+Dot(diff,diff)));
}
//---------------------------------------------------------------------------
double MinLineRectoid (const Line3& line, const Rectoid3& rect,
					   double& r, double& s, double& t)
{
	Point3 diff = { rect.b.x-line.b.x, rect.b.y-line.b.y, rect.b.z-line.b.z };
	double A00 = +Dot(line.m,line.m);
	double A01 = -Dot(line.m,rect.e0);
	double A02 = -Dot(line.m,rect.e1);
	double A11 = +Dot(rect.e0,rect.e0);
	double A12 = +Dot(rect.e0,rect.e1);
	double A22 = +Dot(rect.e1,rect.e1);
	double B0  = -Dot(diff,line.m);
	double B1  = +Dot(diff,rect.e0);
	double B2  = +Dot(diff,rect.e1);
	double cof00 = A11*A22-A12*A12;
	double cof01 = A02*A12-A01*A22;
	double cof02 = A01*A12-A02*A11;
	double det = A00*cof00+A01*cof01+A02*cof02;

	const double tolerance = 1e-05;
	Line3 rectline;
	Point3 pt;
	double min, min0, r0, s0, t0;

	if ( fabs(det) >= tolerance )
	{
		double cof11 = A00*A22-A02*A02;
		double cof12 = A02*A01-A00*A12;
		double cof22 = A00*A11-A01*A01;
		double rhs0 = -B0/det;
		double rhs1 = -B1/det;
		double rhs2 = -B2/det;

		r = cof00*rhs0+cof01*rhs1+cof02*rhs2;
		s = cof01*rhs0+cof11*rhs1+cof12*rhs2;
		t = cof02*rhs0+cof12*rhs1+cof22*rhs2;

		if ( r < 0.0 )
		{
			if ( s < 0.0 )
			{
				if ( t < 0.0 )  // region 6m
				{
					// min on face s=0 or t=0 or r=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t <= 1.0 )  // region 5m
				{
					// min on face s=0 or r=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 4m
				{
					// min on face s=0 or t=1 or r=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
			else if ( s <= 1.0 )
			{
				if ( t < 0.0 )  // region 7m
				{
					// min on face t=0 or r=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min = MinLineLine(line,rectline,r,s);
					t = 0.0;
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t <= 1.0 )  // region 0m
				{
					// min on face r=0
					min = MinPointRectoid(line.b,rect,s,t);
					r = 0.0;
				}
				else  // region 3m
				{
					// min on face t=1 or r=0
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min = MinLineLine(line,rectline,r,s);
					t = 1.0;
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
			else
			{
				if ( t < 0.0 )  // region 8m
				{
					// min on face s=1 or t=0 or r=0
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t <= 1.0 )  // region 1m
				{
					// min on face s=1 or r=0
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 2m
				{
					// min on face s=1 or t=1 or r=0
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					r0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
		}
		else if ( r <= 1.0 )
		{
			if ( s < 0.0 )
			{
				if ( t < 0.0 )  // region 6
				{
					// min on face s=0 or t=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t <= 1.0 )  // region 5
				{
					// min on face s=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
				}
				else // region 4
				{
					// min on face s=0 or t=1
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
			else if ( s <= 1.0 )
			{
				if ( t < 0.0 )  // region 7
				{
					// min on face t=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min = MinLineLine(line,rectline,r,s);
					t = 0.0;
				}
				else if ( t <= 1.0 )  // region 0
				{
					// global minimum is interior
					min = sqrt(fabs(r*(A00*r+A01*s+A02*t+2.0*B0)
						  +s*(A01*r+A11*s+A12*t+2.0*B1)
						  +t*(A02*r+A12*s+A22*t+2.0*B2)
						  +Dot(diff,diff)));
				}
				else  // region 3
				{
					// min on face t=1
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min = MinLineLine(line,rectline,r,s);
					t = 1.0;
				}
			}
			else
			{
				if ( t < 0.0 )  // region 8
				{
					// min on face s=1 or t=0
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t <= 1.0 )  // region 1
				{
					// min on face s=1
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
				}
				else  // region 2
				{
					// min on face s=1 or t=1
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
		}
		else
		{
			if ( s < 0.0 )
			{
				if ( t < 0.0 )  // region 6p
				{
					// min on face s=0 or t=0 or r=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointRectoid(pt,rect,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t <= 1.0 )  // region 5p
				{
					// min on face s=0 or r=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
					min0 = MinPointRectoid(line.b,rect,s0,t0);
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointRectoid(pt,rect,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 4p
				{
					// min on face s=0 or t=1 or r=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 0.0;
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointRectoid(pt,rect,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
			else if ( s <= 1.0 )
			{
				if ( t < 0.0 )  // region 7p
				{
					// min on face t=0 or r=0
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min = MinLineLine(line,rectline,r,s);
					t = 0.0;
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointRectoid(pt,rect,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t <= 1.0 )  // region 0p
				{
					// min on face r=0
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min = MinPointRectoid(pt,rect,s,t);
					r = 0.0;
				}
				else  // region 3p
				{
					// min on face t=1 or r=0
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min = MinLineLine(line,rectline,r,s);
					t = 1.0;
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointRectoid(pt,rect,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
			else
			{
				if ( t < 0.0 )  // region 8p
				{
					// min on face s=1 or t=0 or r=0
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
					rectline.b.x = rect.b.x;
					rectline.b.y = rect.b.y;
					rectline.b.z = rect.b.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 0.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointRectoid(pt,rect,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else if ( t <= 1.0 )  // region 1p
				{
					// min on face s=1 or r=0
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointRectoid(pt,rect,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
				else  // region 2p
				{
					// min on face s=1 or t=1 or r=0
					rectline.b.x = rect.b.x+rect.e0.x;
					rectline.b.y = rect.b.y+rect.e0.y;
					rectline.b.z = rect.b.z+rect.e0.z;
					rectline.m.x = rect.e1.x;
					rectline.m.y = rect.e1.y;
					rectline.m.z = rect.e1.z;
					min = MinLineLine(line,rectline,r,t);
					s = 1.0;
					rectline.b.x = rect.b.x+rect.e1.x;
					rectline.b.y = rect.b.y+rect.e1.y;
					rectline.b.z = rect.b.z+rect.e1.z;
					rectline.m.x = rect.e0.x;
					rectline.m.y = rect.e0.y;
					rectline.m.z = rect.e0.z;
					min0 = MinLineLine(line,rectline,r0,s0);
					t0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
					pt.x = line.b.x+line.m.x;
					pt.y = line.b.y+line.m.y;
					pt.z = line.b.z+line.m.z;
					min0 = MinPointRectoid(pt,rect,s0,t0);
					r0 = 1.0;
					if ( min0 < min )
					{
						min = min0;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
		}
	}
	else
	{
		// line and rectoid are parallel
		rectline.b.x = rect.b.x;
		rectline.b.y = rect.b.y;
		rectline.b.z = rect.b.z;
		rectline.m.x = rect.e0.x;
		rectline.m.y = rect.e0.y;
		rectline.m.z = rect.e0.z;
		min = MinLineLine(line,rectline,r,s);
		t = 0.0;

		rectline.m.x = rect.e1.x;
		rectline.m.y = rect.e1.y;
		rectline.m.z = rect.e1.z;
		min0 = MinLineLine(line,rectline,r0,t0);
		s0 = 0.0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}

		rectline.b.x = rect.b.x+rect.e1.x;
		rectline.b.y = rect.b.y+rect.e1.y;
		rectline.b.z = rect.b.z+rect.e1.z;
		rectline.m.x = rect.e0.x;
		rectline.m.y = rect.e0.y;
		rectline.m.z = rect.e0.z;
		min0 = MinLineLine(line,rectline,r0,s0);
		t0 = 1.0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}

		rectline.b.x = rect.b.x+rect.e0.x;
		rectline.b.y = rect.b.y+rect.e0.y;
		rectline.b.z = rect.b.z+rect.e0.z;
		rectline.m.x = rect.e1.x;
		rectline.m.y = rect.e1.y;
		rectline.m.z = rect.e1.z;
		min0 = MinLineLine(line,rectline,r0,t0);
		s0 = 1.0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}

		min0 = MinPointRectoid(line.b,rect,s0,t0);
		r0 = 0.0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}

		pt.x = line.b.x+line.m.x;
		pt.y = line.b.y+line.m.y;
		pt.z = line.b.z+line.m.z;
		min0 = MinPointRectoid(pt,rect,s0,t0);
		r0 = 1.0;
		if ( min0 < min )
		{
			min = min0;
			r = r0;
			s = s0;
			t = t0;
		}
	}

	return min;
}
//---------------------------------------------------------------------------
double MinTriangleRectoid (const Triangle3& tri, const Rectoid3& rect,
						   double& s, double& t, double& u, double& v)
{
	double s0, t0;  // triangle parameters
	double u0, v0;  // rectangle parameters
	double min, min0;
	Line3 line;

	// compare edges of tri against all of rect
	line.b.x = tri.b.x;
	line.b.y = tri.b.y;
	line.b.z = tri.b.z;
	line.m.x = tri.e0.x;
	line.m.y = tri.e0.y;
	line.m.z = tri.e0.z;
	min = MinLineRectoid(line,rect,s,u,v);
	t = 0.0;

	line.m.x = tri.e1.x;
	line.m.y = tri.e1.y;
	line.m.z = tri.e1.z;
	min0 = MinLineRectoid(line,rect,t0,u0,v0);
	s0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x += tri.e0.x;
	line.b.y += tri.e0.y;
	line.b.z += tri.e0.z;
	line.m.x -= tri.e0.x;
	line.m.y -= tri.e0.y;
	line.m.z -= tri.e0.z;
	min0 = MinLineRectoid(line,rect,t0,u0,v0);  // fix:  replace s0 with t0
	s0 = 1.0-t0;  // fix:  swap s0 and t0
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	// compare edges of rect against all of tri
	line.b.x = rect.b.x;
	line.b.y = rect.b.y;
	line.b.z = rect.b.z;
	line.m.x = rect.e0.x;
	line.m.y = rect.e0.y;
	line.m.z = rect.e0.z;
	min0 = MinLineTriangle(line,tri,u0,s0,t0);
	v0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.m.x = rect.e1.x;
	line.m.y = rect.e1.y;
	line.m.z = rect.e1.z;
	min0 = MinLineTriangle(line,tri,v0,s0,t0);
	u0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x = rect.b.x + rect.e1.x;
	line.b.y = rect.b.y + rect.e1.y;
	line.b.z = rect.b.z + rect.e1.z;
	line.m.x = rect.e0.x;
	line.m.y = rect.e0.y;
	line.m.z = rect.e0.z;
	min0 = MinLineTriangle(line,tri,u0,s0,t0);
	v0 = 1.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x = rect.b.x + rect.e0.x;
	line.b.y = rect.b.y + rect.e0.y;
	line.b.z = rect.b.z + rect.e0.z;
	line.m.x = rect.e1.x;
	line.m.y = rect.e1.y;
	line.m.z = rect.e1.z;
	min0 = MinLineTriangle(line,tri,v0,s0,t0);
	u0 = 1.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	return min;
}
//---------------------------------------------------------------------------
double MinRectoidRectoid (const Rectoid3& rect0, const Rectoid3& rect1,
						  double& s, double& t, double& u, double& v)
{
#if 0
	Triangle3 tri0;
	tri0.b.x = rect0.b.x;
	tri0.b.y = rect0.b.y;
	tri0.b.z = rect0.b.z;
	tri0.e0.x = rect0.e0.x;
	tri0.e0.y = rect0.e0.y;
	tri0.e0.z = rect0.e0.z;
	tri0.e1.x = rect0.e1.x;
	tri0.e1.y = rect0.e1.y;
	tri0.e1.z = rect0.e1.z;

	Triangle3 tri1;
	tri1.b.x = rect0.b.x+rect0.e0.x+rect0.e1.x;
	tri1.b.y = rect0.b.y+rect0.e0.y+rect0.e1.y;
	tri1.b.z = rect0.b.z+rect0.e0.z+rect0.e1.z;
	tri1.e0.x = -rect0.e0.x;
	tri1.e0.y = -rect0.e0.y;
	tri1.e0.z = -rect0.e0.z;
	tri1.e1.x = -rect0.e1.x;
	tri1.e1.y = -rect0.e1.y;
	tri1.e1.z = -rect0.e1.z;

	Triangle3 tri2;
	tri2.b.x = rect1.b.x;
	tri2.b.y = rect1.b.y;
	tri2.b.z = rect1.b.z;
	tri2.e0.x = rect1.e0.x;
	tri2.e0.y = rect1.e0.y;
	tri2.e0.z = rect1.e0.z;
	tri2.e1.x = rect1.e1.x;
	tri2.e1.y = rect1.e1.y;
	tri2.e1.z = rect1.e1.z;

	Triangle3 tri3;
	tri3.b.x = rect1.b.x+rect1.e0.x+rect1.e1.x;
	tri3.b.y = rect1.b.y+rect1.e0.y+rect1.e1.y;
	tri3.b.z = rect1.b.z+rect1.e0.z+rect1.e1.z;
	tri3.e0.x = -rect1.e0.x;
	tri3.e0.y = -rect1.e0.y;
	tri3.e0.z = -rect1.e0.z;
	tri3.e1.x = -rect1.e1.x;
	tri3.e1.y = -rect1.e1.y;
	tri3.e1.z = -rect1.e1.z;

	double min, min0, s0, t0, u0, v0;

	min = MinTriangleTriangle(tri0,tri2,s,t,u,v);

	min0 = MinTriangleTriangle(tri0,tri3,s0,t0,u0,v0);
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	min0 = MinTriangleTriangle(tri1,tri2,s0,t0,u0,v0);
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	min0 = MinTriangleTriangle(tri1,tri3,s0,t0,u0,v0);
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	return min;
#endif
	double s0, t0;  // rect0 parameters
	double u0, v0;  // rect1 parameters
	double min, min0;
	Line3 line;

	// compare edges of rect0 against all of rect1
	line.b.x = rect0.b.x;
	line.b.y = rect0.b.y;
	line.b.z = rect0.b.z;
	line.m.x = rect0.e0.x;
	line.m.y = rect0.e0.y;
	line.m.z = rect0.e0.z;
	min = MinLineRectoid(line,rect1,s,u,v);
	t = 0.0;

	line.m.x = rect0.e1.x;
	line.m.y = rect0.e1.y;
	line.m.z = rect0.e1.z;
	min0 = MinLineRectoid(line,rect1,t0,u0,v0);
	s0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x = rect0.b.x + rect0.e0.x;
	line.b.y = rect0.b.y + rect0.e0.y;
	line.b.z = rect0.b.z + rect0.e0.z;
	min0 = MinLineRectoid(line,rect1,t0,u0,v0);
	s0 = 1.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x = rect0.b.x + rect0.e1.x;
	line.b.y = rect0.b.y + rect0.e1.y;
	line.b.z = rect0.b.z + rect0.e1.z;
	line.m.x = rect0.e0.x;
	line.m.y = rect0.e0.y;
	line.m.z = rect0.e0.z;
	min0 = MinLineRectoid(line,rect1,s0,u0,v0);
	t0 = 1.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	// compare edges of rect1 against all of rect0
	line.b.x = rect1.b.x;
	line.b.y = rect1.b.y;
	line.b.z = rect1.b.z;
	line.m.x = rect1.e0.x;
	line.m.y = rect1.e0.y;
	line.m.z = rect1.e0.z;
	min0 = MinLineRectoid(line,rect0,u0,s0,t0);
	v0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.m.x = rect1.e1.x;
	line.m.y = rect1.e1.y;
	line.m.z = rect1.e1.z;
	min0 = MinLineRectoid(line,rect0,v0,s0,t0);
	u0 = 0.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x = rect1.b.x + rect1.e0.x;
	line.b.y = rect1.b.y + rect1.e0.y;
	line.b.z = rect1.b.z + rect1.e0.z;
	min0 = MinLineRectoid(line,rect0,v0,s0,t0);
	u0 = 1.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	line.b.x = rect1.b.x + rect1.e0.x;
	line.b.y = rect1.b.y + rect1.e0.y;
	line.b.z = rect1.b.z + rect1.e0.z;
	line.m.x = rect1.e0.x;
	line.m.y = rect1.e0.y;
	line.m.z = rect1.e0.z;
	min0 = MinLineRectoid(line,rect0,u0,s0,t0);
	v0 = 1.0;
	if ( min0 < min )
	{
		min = min0;
		s = s0;
		t = t0;
		u = u0;
		v = v0;
	}

	return min;
}
//---------------------------------------------------------------------------

#ifdef GEOMDIST_TEST

#include <iostream.h>
#include <float.h>

#if 1

//#define LINE_LINE_TEST
#ifdef LINE_LINE_TEST

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

ofstream ostr("data.txt");

#define RAND (2.0*rand()/double(RAND_MAX)-1.0)

void main ()
{
	Line3 line0, line1;
	Point3 p0, p1, diff;
	double s, t, s0, t0, min, dist;

	for (int i = 0; i < 128; i++)
	{
		line0.b.x = RAND;
		line0.b.y = RAND;
		line0.b.z = RAND;

		line0.m.x = RAND;
		line0.m.y = RAND;
		line0.m.z = RAND;

		line1.b.x = RAND;
		line1.b.y = RAND;
		line1.b.z = RAND;

		if ( i % 2 )
		{
			// non-parallel line segments
			line1.m.x = RAND;
			line1.m.y = RAND;
			line1.m.z = RAND;
		}
		else
		{
			// parallel line segments
			double scale = RAND;
			line1.m.x = scale*line0.m.x;
			line1.m.y = scale*line0.m.y;
			line1.m.z = scale*line0.m.z;
		}

		min = DBL_MAX;
		int ymax = 128, xmax = 128;
		for (int y = 0; y < ymax; y++)
		{
			s0 = y/double(ymax-1);
			p0.x = line0.b.x+s0*line0.m.x;
			p0.y = line0.b.y+s0*line0.m.y;
			p0.z = line0.b.z+s0*line0.m.z;
			for (int x = 0; x < xmax; x++)
			{
				t0 = x/double(xmax-1);
				p1.x = line1.b.x+t0*line1.m.x;
				p1.y = line1.b.y+t0*line1.m.y;
				p1.z = line1.b.z+t0*line1.m.z;

				diff.x = p1.x-p0.x;
				diff.y = p1.y-p0.y;
				diff.z = p1.z-p0.z;
				dist = sqrt(Dot(diff,diff));
				if ( dist < min )
				{
					min = dist;
					s = s0;
					t = t0;
				}
			}
		}
		ostr << "actual = " << s << ' ' << t << ' ' << min << endl;

		min = MinLineLine(line0,line1,s,t);
		ostr << "approx = " << s << ' ' << t << ' ' << min << endl;

		ostr << endl;
	}
}
#endif

//#define POINT_TRIANGLE_TEST
#ifdef POINT_TRIANGLE_TEST

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

ofstream ostr("data.txt");

#define RAND (2.0*rand()/double(RAND_MAX)-1.0)

void main ()
{
	Triangle3 tri;
	Point3 p, q, diff;
	double s, t, s0, t0, min0, min1, dist;

	double maxdiff = 0.0;

	for (int i = 0; i < 128; i++)
	{
		tri.b.x = RAND;
		tri.b.y = RAND;
		tri.b.z = RAND;
		tri.e0.x = RAND;
		tri.e0.y = RAND;
		tri.e0.z = RAND;
		tri.e1.x = RAND;
		tri.e1.y = RAND;
		tri.e1.z = RAND;

		p.x = RAND;
		p.y = RAND;
		p.z = RAND;

		min0 = DBL_MAX;
		int max = 128;
		for (int y = 0; y <= max; y++)
		{
			s0 = y/double(max);
			for (int x = 0; x+y <= max; x++)
			{
				t0 = x/double(max);
				q.x = tri.b.x+s0*tri.e0.x+t0*tri.e1.x;
				q.y = tri.b.y+s0*tri.e0.y+t0*tri.e1.y;
				q.z = tri.b.z+s0*tri.e0.z+t0*tri.e1.z;

				diff.x = p.x-q.x;
				diff.y = p.y-q.y;
				diff.z = p.z-q.z;
				dist = sqrt(Dot(diff,diff));
				if ( dist < min0 )
				{
					min0 = dist;
					s = s0;
					t = t0;
				}
			}
		}
		ostr << "actual = " << s << ' ' << t << ' ' << min0 << endl;

		min1 = MinPointTriangle(p,tri,s,t);
		ostr << "approx = " << s << ' ' << t << ' ' << min1 << endl;

		ostr << "diff = " << min1-min0 << endl;

		if ( min1-min0 > maxdiff )
			maxdiff = min1-min0;

		ostr << endl;
	}
	ostr << "max diff = " << maxdiff << endl;
}
#endif

//#define LINE_TRIANGLE_TEST
#ifdef LINE_TRIANGLE_TEST

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

ofstream ostr("data.txt");

#define RAND (2.0*rand()/double(RAND_MAX)-1.0)

void main ()
{
	Triangle3 tri;
	Line3 line;
	Point3 p, q, diff;
	double r, s, t, r0, s0, t0, min0, min1, dist;

	double maxdiff = 0.0;

	for (int i = 0; i < 128; i++)
	{
		tri.b.x = RAND;
		tri.b.y = RAND;
		tri.b.z = RAND;
		tri.e0.x = RAND;
		tri.e0.y = RAND;
		tri.e0.z = RAND;
		tri.e1.x = RAND;
		tri.e1.y = RAND;
		tri.e1.z = RAND;

		line.b.x = RAND;
		line.b.y = RAND;
		line.b.z = RAND;
		if ( i % 2 )
		{
			// non-parallel line and triangle
			line.m.x = RAND;
			line.m.y = RAND;
			line.m.z = RAND;
		}
		else
		{
			// line is parallel to triangle
			double c0 = RAND;
			double c1 = RAND;
			line.m.x = c0*tri.e0.x+c1*tri.e1.x;
			line.m.y = c0*tri.e0.y+c1*tri.e1.y;
			line.m.z = c0*tri.e0.z+c1*tri.e1.z;
		}

		min0 = DBL_MAX;
		int max = 32;
		for (int z = 0; z <= max; z++)
		{
			r0 = z/double(max);
			p.x = line.b.x+r0*line.m.x;
			p.y = line.b.y+r0*line.m.y;
			p.z = line.b.z+r0*line.m.z;
			for (int y = 0; y <= max; y++)
			{
				s0 = y/double(max);
				for (int x = 0; x+y <= max; x++)
				{
					t0 = x/double(max);
					q.x = tri.b.x+s0*tri.e0.x+t0*tri.e1.x;
					q.y = tri.b.y+s0*tri.e0.y+t0*tri.e1.y;
					q.z = tri.b.z+s0*tri.e0.z+t0*tri.e1.z;
					
					diff.x = p.x-q.x;
					diff.y = p.y-q.y;
					diff.z = p.z-q.z;
					dist = sqrt(Dot(diff,diff));
					if ( dist < min0 )
					{
						min0 = dist;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
		}
		ostr << "i = " << i << endl;
		ostr << "actual = " << r << ' ' << s << ' ' << t << ' ' << min0 << endl;

		min1 = MinLineTriangle(line,tri,r,s,t);
		ostr << "approx = " << r << ' ' << s << ' ' << t << ' ' << min1 << endl;

		ostr << "diff = " << min1-min0 << endl;

		if ( min1-min0 > maxdiff )
			maxdiff = min1-min0;

		ostr << endl;
	}

	ostr << "max diff = " << maxdiff << endl;
}
#endif

//#define TRIANGLE_TRIANGLE_TEST
#ifdef TRIANGLE_TRIANGLE_TEST

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

ofstream ostr("data.txt");

#define RAND (2.0*rand()/double(RAND_MAX)-1.0)

void main ()
{
	Triangle3 tri0, tri1;
	Point3 p, q, diff;
	double u, v, s, t, u0, v0, s0, t0, min0, min1, dist;

	double maxdiff = 0.0;

	for (int i = 0; i < 128; i++)
	{
		tri0.b.x = RAND;
		tri0.b.y = RAND;
		tri0.b.z = RAND;
		tri0.e0.x = RAND;
		tri0.e0.y = RAND;
		tri0.e0.z = RAND;
		tri0.e1.x = RAND;
		tri0.e1.y = RAND;
		tri0.e1.z = RAND;

		tri1.b.x = RAND;
		tri1.b.y = RAND;
		tri1.b.z = RAND;
		tri1.e0.x = RAND;
		tri1.e0.y = RAND;
		tri1.e0.z = RAND;
		tri1.e1.x = RAND;
		tri1.e1.y = RAND;
		tri1.e1.z = RAND;

		min0 = DBL_MAX;
		int max = 32;
		for (int w = 0; w <= max; w++)
		{
			t0 = w/double(max);
			for (int z = 0; w+z <= max; z++)
			{
				s0 = z/double(max);
				p.x = tri0.b.x+s0*tri0.e0.x+t0*tri0.e1.x;
				p.y = tri0.b.y+s0*tri0.e0.y+t0*tri0.e1.y;
				p.z = tri0.b.z+s0*tri0.e0.z+t0*tri0.e1.z;
				for (int y = 0; y <= max; y++)
				{
					v0 = y/double(max);
					for (int x = 0; x+y <= max; x++)
					{
						u0 = x/double(max);
						q.x = tri1.b.x+u0*tri1.e0.x+v0*tri1.e1.x;
						q.y = tri1.b.y+u0*tri1.e0.y+v0*tri1.e1.y;
						q.z = tri1.b.z+u0*tri1.e0.z+v0*tri1.e1.z;
						
						diff.x = p.x-q.x;
						diff.y = p.y-q.y;
						diff.z = p.z-q.z;
						dist = sqrt(Dot(diff,diff));
						if ( dist < min0 )
						{
							min0 = dist;
							s = s0;
							t = t0;
							u = u0;
							v = v0;
						}
					}
				}
			}
		}
		ostr << "i = " << i << endl;
		ostr << "actual = " << s << ' ' << t << ' ' << u << ' ' << v << ' ' << min0 << endl;

		min1 = MinTriangleTriangle(tri0,tri1,s,t,u,v);
		ostr << "approx = " << s << ' ' << t << ' ' << u << ' ' << v << ' ' << min1 << endl;

		ostr << "diff = " << min1-min0 << endl;

		if ( min1-min0 > maxdiff )
			maxdiff = min1-min0;

		ostr << endl;
	}

	ostr << "max diff = " << maxdiff << endl;
}
#endif

//#define POINT_RECTANGLE_TEST
#ifdef POINT_RECTANGLE_TEST

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

ofstream ostr("data.txt");

#define RAND (2.0*rand()/double(RAND_MAX)-1.0)

void main ()
{
	Rectangle3 rect;
	Point3 p, q, diff;
	double s, t, s0, t0, min0, min1, dist;
	double maxdiff = 0.0;

	for (int i = 0; i < 128; i++)
	{
		rect.b.x = RAND;
		rect.b.y = RAND;
		rect.b.z = RAND;
		rect.e0.x = RAND;
		rect.e0.y = RAND;
		rect.e0.z = RAND;
		rect.e1.x = RAND;
		rect.e1.y = RAND;
		rect.e1.z = RAND;

		p.x = RAND;
		p.y = RAND;
		p.z = RAND;

		min0 = DBL_MAX;
		int max = 128;
		for (int y = 0; y <= max; y++)
		{
			s0 = y/double(max);
			for (int x = 0; x <= max; x++)
			{
				t0 = x/double(max);
				q.x = rect.b.x+s0*rect.e0.x+t0*rect.e1.x;
				q.y = rect.b.y+s0*rect.e0.y+t0*rect.e1.y;
				q.z = rect.b.z+s0*rect.e0.z+t0*rect.e1.z;

				diff.x = p.x-q.x;
				diff.y = p.y-q.y;
				diff.z = p.z-q.z;
				dist = sqrt(Dot(diff,diff));
				if ( dist < min0 )
				{
					min0 = dist;
					s = s0;
					t = t0;
				}
			}
		}
		ostr << "actual = " << s << ' ' << t << ' ' << min0 << endl;

		min1 = MinPointRectangle(p,rect,s,t);
		ostr << "approx = " << s << ' ' << t << ' ' << min1 << endl;
		ostr << "diff = " << min1-min0 << endl;
		if ( min1-min0 > maxdiff )
			maxdiff = min1-min0;
		ostr << endl;
	}

	ostr << "maxdiff = " << maxdiff << endl;
	ostr << endl;
}
#endif

//#define LINE_RECTANGLE_TEST
#ifdef LINE_RECTANGLE_TEST

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

ofstream ostr("data.txt");

#define RAND (2.0*rand()/double(RAND_MAX)-1.0)

void main ()
{
	Rectoid3 rect;
	Line3 line;
	Point3 p, q, diff;
	double r, s, t, r0, s0, t0, min0, min1, dist;

	double maxdiff = 0.0;

	for (int i = 0; i < 128; i++)
	{
		rect.b.x = RAND;
		rect.b.y = RAND;
		rect.b.z = RAND;
		rect.e0.x = RAND;
		rect.e0.y = RAND;
		rect.e0.z = RAND;
		rect.e1.x = RAND;
		rect.e1.y = RAND;
		rect.e1.z = RAND;

		line.b.x = RAND;
		line.b.y = RAND;
		line.b.z = RAND;
		if ( i % 2 )
		{
			// non-parallel line and rectoid
			line.m.x = RAND;
			line.m.y = RAND;
			line.m.z = RAND;
		}
		else
		{
			// line is parallel to rectoid
			double c0 = RAND;
			double c1 = RAND;
			line.m.x = c0*rect.e0.x+c1*rect.e1.x;
			line.m.y = c0*rect.e0.y+c1*rect.e1.y;
			line.m.z = c0*rect.e0.z+c1*rect.e1.z;
		}

		min0 = DBL_MAX;
		int max = 32;
		for (int z = 0; z <= max; z++)
		{
			r0 = z/double(max);
			p.x = line.b.x+r0*line.m.x;
			p.y = line.b.y+r0*line.m.y;
			p.z = line.b.z+r0*line.m.z;
			for (int y = 0; y <= max; y++)
			{
				s0 = y/double(max);
				for (int x = 0; x <= max; x++)
				{
					t0 = x/double(max);
					q.x = rect.b.x+s0*rect.e0.x+t0*rect.e1.x;
					q.y = rect.b.y+s0*rect.e0.y+t0*rect.e1.y;
					q.z = rect.b.z+s0*rect.e0.z+t0*rect.e1.z;
					
					diff.x = p.x-q.x;
					diff.y = p.y-q.y;
					diff.z = p.z-q.z;
					dist = sqrt(Dot(diff,diff));
					if ( dist < min0 )
					{
						min0 = dist;
						r = r0;
						s = s0;
						t = t0;
					}
				}
			}
		}
		ostr << "i = " << i << endl;
		ostr << "actual = " << r << ' ' << s << ' ' << t << ' ' << min0 << endl;

		min1 = MinLineRectoid(line,rect,r,s,t);
		ostr << "approx = " << r << ' ' << s << ' ' << t << ' ' << min1 << endl;
		ostr << "diff = " << min1-min0 << endl;

		if ( min1-min0 > maxdiff )
			maxdiff = min1-min0;

		ostr << endl;
	}

	ostr << "max diff = " << maxdiff << endl;
}
#endif

//#define TRIANGLE_RECTANGLE_TEST
#ifdef TRIANGLE_RECTANGLE_TEST

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

ofstream ostr("data.txt");

#define RAND (2.0*rand()/double(RAND_MAX)-1.0)

void main ()
{
	Triangle3 tri;
	Rectoid3 rect;
	Point3 p, q, diff;
	double u, v, s, t, u0, v0, s0, t0, min0, min1, dist;

	double maxdiff = 0.0;

	for (int i = 0; i < 128; i++)
	{
		tri.b.x = RAND;
		tri.b.y = RAND;
		tri.b.z = RAND;
		tri.e0.x = RAND;
		tri.e0.y = RAND;
		tri.e0.z = RAND;
		tri.e1.x = RAND;
		tri.e1.y = RAND;
		tri.e1.z = RAND;

		rect.b.x = RAND;
		rect.b.y = RAND;
		rect.b.z = RAND;
		rect.e0.x = RAND;
		rect.e0.y = RAND;
		rect.e0.z = RAND;
		rect.e1.x = RAND;
		rect.e1.y = RAND;
		rect.e1.z = RAND;

		min0 = DBL_MAX;
		int max = 32;
		for (int w = 0; w <= max; w++)
		{
			t0 = w/double(max);
			for (int z = 0; w+z <= max; z++)
			{
				s0 = z/double(max);
				p.x = tri.b.x+s0*tri.e0.x+t0*tri.e1.x;
				p.y = tri.b.y+s0*tri.e0.y+t0*tri.e1.y;
				p.z = tri.b.z+s0*tri.e0.z+t0*tri.e1.z;
				for (int y = 0; y <= max; y++)
				{
					v0 = y/double(max);
					for (int x = 0; x <= max; x++)
					{
						u0 = x/double(max);
						q.x = rect.b.x+u0*rect.e0.x+v0*rect.e1.x;
						q.y = rect.b.y+u0*rect.e0.y+v0*rect.e1.y;
						q.z = rect.b.z+u0*rect.e0.z+v0*rect.e1.z;
						
						diff.x = p.x-q.x;
						diff.y = p.y-q.y;
						diff.z = p.z-q.z;
						dist = sqrt(Dot(diff,diff));
						if ( dist < min0 )
						{
							min0 = dist;
							s = s0;
							t = t0;
							u = u0;
							v = v0;
						}
					}
				}
			}
		}
		ostr << "i = " << i << endl;
		ostr << "actual = " << s << ' ' << t << ' ' << u << ' ' << v << ' ' << min0 << endl;

		min1 = MinTriangleRectoid(tri,rect,s,t,u,v);
		ostr << "approx = " << s << ' ' << t << ' ' << u << ' ' << v << ' ' << min1 << endl;

		ostr << "diff = " << min1-min0 << endl;

		if ( min1-min0 > maxdiff )
			maxdiff = min1-min0;

		ostr << endl;
	}

	ostr << "max diff = " << maxdiff << endl;
}
#endif

#define RECTANGLE_RECTANGLE_TEST
#ifdef RECTANGLE_RECTANGLE_TEST

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

ofstream ostr("data.txt");

#define RAND (2.0*rand()/double(RAND_MAX)-1.0)

void main ()
{
	Rectoid3 rect0, rect1;
	Point3 p, q, diff;
	double u, v, s, t, u0, v0, s0, t0, min0, min1, dist;

	double maxdiff = 0.0;

	for (int i = 0; i < 128; i++)
	{
		rect0.b.x = RAND;
		rect0.b.y = RAND;
		rect0.b.z = RAND;
		rect0.e0.x = RAND;
		rect0.e0.y = RAND;
		rect0.e0.z = RAND;
		rect0.e1.x = RAND;
		rect0.e1.y = RAND;
		rect0.e1.z = RAND;

		rect1.b.x = RAND;
		rect1.b.y = RAND;
		rect1.b.z = RAND;
		rect1.e0.x = RAND;
		rect1.e0.y = RAND;
		rect1.e0.z = RAND;
		rect1.e1.x = RAND;
		rect1.e1.y = RAND;
		rect1.e1.z = RAND;

		min0 = DBL_MAX;
		int max = 32;
		for (int w = 0; w <= max; w++)
		{
			t0 = w/double(max);
			for (int z = 0; z <= max; z++)
			{
				s0 = z/double(max);
				p.x = rect0.b.x+s0*rect0.e0.x+t0*rect0.e1.x;
				p.y = rect0.b.y+s0*rect0.e0.y+t0*rect0.e1.y;
				p.z = rect0.b.z+s0*rect0.e0.z+t0*rect0.e1.z;
				for (int y = 0; y <= max; y++)
				{
					v0 = y/double(max);
					for (int x = 0; x <= max; x++)
					{
						u0 = x/double(max);
						q.x = rect1.b.x+u0*rect1.e0.x+v0*rect1.e1.x;
						q.y = rect1.b.y+u0*rect1.e0.y+v0*rect1.e1.y;
						q.z = rect1.b.z+u0*rect1.e0.z+v0*rect1.e1.z;
						
						diff.x = p.x-q.x;
						diff.y = p.y-q.y;
						diff.z = p.z-q.z;
						dist = sqrt(Dot(diff,diff));
						if ( dist < min0 )
						{
							min0 = dist;
							s = s0;
							t = t0;
							u = u0;
							v = v0;
						}
					}
				}
			}
		}
		ostr << "i = " << i << endl;
		ostr << "actual = " << s << ' ' << t << ' ' << u << ' ' << v << ' ' << min0 << endl;

		min1 = MinRectoidRectoid(rect0,rect1,s,t,u,v);
		ostr << "approx = " << s << ' ' << t << ' ' << u << ' ' << v << ' ' << min1 << endl;

		ostr << "diff = " << min1-min0 << endl;

		if ( min1-min0 > maxdiff )
			maxdiff = min1-min0;

		ostr << endl;
	}

	ostr << "max diff = " << maxdiff << endl;
}
#endif
#endif

#endif
